#!/usr/local/bin/python3
import sys,math,subprocess

# ph2vesta.py
# This is a post-processing program for Quantum-Espresso's ph.x. 
# After the normal vibration mode calculation, it adds the atomic displacements 
# of each vibration mode as arrows (vectors) to the atoms in VESTA's 
# crystal structure display.
# made by mkanzaki@me.com
# created: 2017 06/23: primitive first version
# modified: 2022 09/15: Updated. Now Python3. Corrections for newer output format.
# bug fix: 2022 12/12: When E or T exists, nmode was not correct
# modified: 2023 11/14: Now original vesta file is necessary for input file and becomes simpler
# modified: 2023 11/16: 'a' option (export all modes) added
#
#
# To run, you need the VESTA file of the structure used for the normal mode 
# calculation (this can be created with pwout2xtl.py), the ph.x output 
# file and the dyn file.

# Usage:
# First, run the pw.x calculation (scf or relax or vc-relax). 
# pwout2xtl.py program is used to create a xtl file, and open by VESTA
# adjust the direction etc., you want to see, and then save it as VESTA file. 
# Next, run the ph.x program to perform the normal mode calculation (q=0). 
# Then run the following.
# > . /ph2vesta.py ph-out-file dyn-file vesta-file
# 1st argument: phonon output file name
# 2nd argument: dyn-file name
# 3rd argument: VESTA file name
# When executed, a list of vibration modes is displayed. Select a mode by number, 
# the corresponding mode with arrows will automatically appear in VESTA.
# Type q to exit.
# Type 0 to change settings such as arrow size and direction.

# Note:
# You can also change arrow length, diameter and color from VESTA's Vectors...
# When run without argument, it will print out very short how-to-use.
# This code is for Mac. First change the first line to match your python path.
# This code is for Mac. First change the first line to match your python path.
# For Windows or Linux, change the last part that calls VESTA. 
# Or, comment out (cmd =... and subprocess.call) and read the produced VESTA file from VESTA.
#
celldm = [0.0 for i in range(7)]
mat1 = [0.0 for i in range(4)]
mat2 = [0.0 for i in range(4)]
mat3 = [0.0 for i in range(4)]
rmat1 = [0.0 for i in range(4)]
rmat2 = [0.0 for i in range(4)]
rmat3 = [0.0 for i in range(4)]

maxmode = 500 # suppose total modes less than 500, but increase if exceeds
atom = ['' for i in range(maxmode)]
freq = [0.0 for i in range(maxmode)]
mode = ['' for i in range(maxmode)]
x = [0.0 for i in range(maxmode)] 
y = [0.0 for i in range(maxmode)] # increase if necessary
z = [0.0 for i in range(maxmode)]
dx = [[0.0 for i in range(maxmode)] for j in range(maxmode)]
dy = [[0.0 for i in range(maxmode)] for j in range(maxmode)]
dz = [[0.0 for i in range(maxmode)] for j in range(maxmode)]

if len(sys.argv) < 4:
	print('Not enough arguments provided!')
	print('For example: > ./ph2vesta.py ph_out_file dyn_file vesta_file')
	print('First argument: ph out file name')
	print('Second argument: dyn file name')
	print('Third argument: vesta file name')
	print('Read header of this code for details.')
	exit()

scale = 2.0 # arrow length (same scale for all), change if necessary
radius = 0.3 # arrow radius (currently same for all),  change if necessary
arrowRGB = '255 0 0' # arrow color RGB (currently same for all) default red
# ----------------- need to change:  magnification factor ---------------
magnification = 1.0
print('Current scale factor for displacement is ',scale)
# reading argument
base = sys.argv[1]

# change below according to your preferred file naming convention 
f1 = sys.argv[1]  # file for reading crystal structure and mode info
f2 = sys.argv[2] # file for reading vibrational displacements
f5 = sys.argv[3] # for reading vesta file
try:
	file1 = open(f1,'r')
except IOError as xxx_todo_changeme:
	(errno, msg) = xxx_todo_changeme.args
	print(f1 + ' file open error!')
	exit()
try:
	file2 = open(f2,'r')
except IOError as xxx_todo_changeme1:
	(errno, msg) = xxx_todo_changeme1.args
	print(f2 + ' file open error!')
	exit()
try:
	file5 = open(f5,'r')
except IOError as xxx_todo_changeme1:
	(errno, msg) = xxx_todo_changeme1.args
	print(f5 + ' file open error!')
	exit()

# define constants
rd = math.pi/180.0
bohr = 0.5291772108 # used to convert from bohr to Ang.

#---subroutine---
def vestaout(qmode): # make vesta file of specified mode
	global f3
	title = base + '-' + mode[qmode].strip() + '-' + str('%04.1f' % freq[qmode]) + 'cm-1' # 
	f3 = title + '.vesta' # file name
	file3 = open(f3,'w') # open original VESTA file
	for i in range(0,ix1): # same as original up to TITLE
		ftext = all_lines5[i]
		if 'TITLE' in ftext:
			break
		file3.write(ftext)
	file3.write(ftext)
	file3.write('ph.x: ' + title + '\n')
	ix0 = i + 2 # skip old title line
	for i in range(ix0,ix1): # same as original up to ix1 line
		file3.write(all_lines5[i])
	# below are Vector (arrow) part. Depends on the mode selected
	file3.write('VECTR\n') # defining arrow directions
	for i in range(1,natom+1):
		t1=str(scale*dx[qmode][i]) # scale is applied here
		t2=str(scale*dy[qmode][i])
		t3=str(scale*dz[qmode][i])
		file3.write(' ' + str(i) + ' ' + t1 + ' ' + t2 + ' ' + t3 + ' 0\n')
		file3.write('   ' + str(i) + '   0    0    0    0\n')
		file3.write(' 0 0 0 0 0\n')
	file3.write(' 0 0 0 0 0\n')
	file3.write('VECTT\n') # defining color and style
	for i in range(1,natom+1):
		# change here for arrow color, style
		# first one: radius of arrow (0.3), color (255 0 0), mode (1 or 2)
		# mode=1: penetrated, mode=2: on radius, currently fixed to 2
		file3.write(' ' + str(i) + '  ' + str(radius) + ' ' + arrowRGB + ' 2\n')
	file3.write(' 0 0 0 0 0\n')
	for i in range(ix2,len(all_lines5)): # after VECTT, same as original VESTA file
		file3.write(all_lines5[i])
	file3.close()

#---subroutine---

all_lines1 = file1.readlines()
file1.close()
findex = 0 # position of current file1 or file2 lines

# Get Bravais index (cell) from *.ph.out
for i in range(0,len(all_lines1)): # find bravais-lattice index line
	ftext = all_lines1[i]
	if 'bravais-lattice index' in ftext :
	   break
s =ftext.split()
ibrav = int(s[3]) # ibrav number integer
findex = i # last i

# Get alat from *.out
ftext = all_lines1[findex + 1]
s =ftext.split()
alat = float(s[4]) # alat (celldm[1])
#print(alat)

# Number of atoms in cell
for i in range(findex,len(all_lines1)): # find bravais-lattice index line
	ftext = all_lines1[i]
	if 'number of atoms/cell' in ftext :
	   break
s =ftext.split()
natom = int(s[4]) # number of atoms in a cell
nmode = natom * 3 # xyz freedom times total number of atoms in the unit cell
findex = i # last i

#read crystal axes (necessary to convert to cell parameters)
for i in range(findex,len(all_lines1)): # find crystal axes: line
	ftext = all_lines1[i]
	if 'crystal axes:' in ftext :
		break
findex = i
ftext = all_lines1[findex+1] # next line
s = ftext.split()
mat1[1] = float(s[3])
mat1[2] = float(s[4])
mat1[3] = float(s[5])
ftext = all_lines1[findex+2] # next line
s = ftext.split()
mat2[1] = float(s[3])
mat2[2] = float(s[4])
mat2[3] = float(s[5])

ftext = all_lines1[findex+3] # next line
s = ftext.split()
mat3[1] = float(s[3])
mat3[2] = float(s[4])
mat3[3] = float(s[5])
findex = findex + 4

#read reciprocal axes (necessary for converting coordinates)
for i in range(findex,len(all_lines1)): # find crystal axes: line
	ftext = all_lines1[i]
	if 'reciprocal axes:' in ftext :
		break
findex = i
ftext = all_lines1[findex+1] # next line
s = ftext.split()
rmat1[1] = float(s[3])
rmat1[2] = float(s[4])
rmat1[3] = float(s[5])

ftext = all_lines1[findex+2] # next line
s = ftext.split()
rmat2[1] = float(s[3])
rmat2[2] = float(s[4])
rmat2[3] = float(s[5])

ftext = all_lines1[findex+3] # next line
s = ftext.split()
rmat3[1] = float(s[3])
rmat3[2] = float(s[4])
rmat3[3] = float(s[5])
findex = findex + 4
#print(rmat1[1],rmat1[2],rmat1[3],rmat2[1],rmat2[2],rmat2[3],rmat3[1],rmat3[2],rmat3[3])

# calculate cell parameters from alat, mat matrix
cella = math.sqrt(mat1[1]*mat1[1]+mat1[2]*mat1[2]+mat1[3]*mat1[3])
cellb = math.sqrt(mat2[1]*mat2[1]+mat2[2]*mat2[2]+mat2[3]*mat2[3])
cellc = math.sqrt(mat3[1]*mat3[1]+mat3[2]*mat3[2]+mat3[3]*mat3[3])
alpha = math.acos((mat2[1]*mat3[1]+mat2[2]*mat3[2]+mat2[3]*mat3[3])/(cellb*cellc))/rd
beta = math.acos((mat1[1]*mat3[1]+mat1[2]*mat3[2]+mat1[3]*mat3[3])/(cella*cellc))/rd
gamma = math.acos((mat1[1]*mat2[1]+mat1[2]*mat2[2]+mat1[3]*mat2[3])/(cella*cellb))/rd
cella = alat*bohr*cella
cellb = alat*bohr*cellb
cellc = alat*bohr*cellc
#print(cella,cellb,cellc,alpha,beta,gamma)

# read atomic coordinates (they are alat units, so conversion is necessary) 
# For conversion, reciprocal axes rmat[] are used 

for i in range(findex,len(all_lines1)): # find site n. atom... line
	ftext = all_lines1[i]
	if 'site n.  atom' in ftext :
		break
findex = i + 1 # point next line

for i in range(0,natom): # find site n. atom... line
	ftext = all_lines1[findex+i]
	s = ftext.split()
	atom[i+1] = s[1]
	x1 = float(s[7])
	y1 = float(s[8])
	z1 = float(s[9])
	x[i+1]= x1*rmat1[1] + y1*rmat1[2] + z1*rmat1[3] # conversion
	y[i+1]= x1*rmat2[1] + y1*rmat2[2] + z1*rmat2[3] # conversion 
	z[i+1]= x1*rmat3[1] + y1*rmat3[2] + z1*rmat3[3] # conversion
#	print(x[i+1],y[i+1],z[i+1])
findex = findex + natom # point current line

# find Mode symmetry part
for i in range(findex,len(all_lines1)): # find first ******* line
	ftext = all_lines1[i]
	if '***************' in ftext :
		break
ij = i
for i in range(ij+1,len(all_lines1)): # find second ******* line
	ftext = all_lines1[i]
	if '***************' in ftext :
		break
findex = i + 1 # last i

# find line contains Mode symmetry
for i in range(findex,len(all_lines1)): # find site n. atom... line
	ftext = all_lines1[i]
	if 'Mode symmetry,' in ftext :
		break
findex = i + 2 # point next^2 line, mode starts here

# obtain number of actual lines for mode symmetry 
# Not equal to nmode if this contains E or T modes and has cneter of symmetry
m = 0
for i in range(findex,findex + nmode): # find site n. atom... line
	ftext = all_lines1[i]
	if 'freq' in ftext :
		m = m + 1
	else:
		break
nmode2 = m
if nmode != nmode2:
	print('Degenerated modes detected.')

# read character representation (B_1g etc.) and vibrational frequency
imode = 1
for i in range(0,nmode2): # read frequency and mode 
	ftext = all_lines1[findex+i]
	s = ftext.split()
	s1 = s[2]
	if s1[len(s1)-1] == '-': # 6.7 or later need this
		s1 = s1.rstrip('-')
		m1 = int(s1)
		f1 = float(s[5])
		str1 = s[8]
		s2 = s[3]
		s2 = s2.rstrip(')')
		m2 = int(s2)	
	else: # previous version
		m1 = int(s1)
		f1 = float(s[6])
		str1 = s[9]
		s2 = s[4]
		s2 = s2.rstrip(')')
		m2 = int(s2)
	#print(m1,m2)
	j = m2 - m1 + 1
	for k in range(1,j+1):
		mode[imode] = str1
		freq[imode] = f1
		imode = imode + 1
imode = imode -1
#print(imode)

# read from *.dyn (file2)
all_lines2 = file2.readlines()
file2.close()
findex = 1
for i in range(findex,len(all_lines2)): # find Diagonalizing the dynamical matrix
	ftext = all_lines2[i]
	if 'Diagonalizing the dynamical matrix' in ftext :
		break
findex = i # last i

for i in range(findex,len(all_lines2)): # find ****** line
	ftext = all_lines2[i]
	if '******' in ftext :
		break
findex = i # last i
if (findex + 1) >= (len(all_lines2)): # if something wrong...
   print('End of line reached!')
   exit(0)
# read displacement vectors dx[mode][atom],dy[mode][atom],dz[mode][atom]
for i in range(1,imode+1): 
	findex = findex + 1
	for j in range(1,natom+1):
		#print(len(all_lines2),findex,j)
		ftext = all_lines2[findex + j]
		s = ftext.split()
		dx[i][j] = float(s[1])
		dy[i][j] = float(s[3])
		dz[i][j] = float(s[5])
	findex = findex + natom
# read from *.dyn (file2) end

# read from VESTA file (file5) new 2023/11/14
all_lines5 = file5.readlines()
for i in range(0,len(all_lines5)):
	ftext = all_lines5[i]
	if 'VECTR' in ftext :
		ix1 = i # pointing a line before VECTR
		ix2 = i + 4 # pointing a line after VECTT 0 0 0 0 0
file5.close()

# Interactively select mode, and make a VESTA file for the selected mode
# Now can export vesta file for all modes by "a" option
while True: # print list of the modes
	for i in range(1,imode+1):
		print(' ' + str(i) + '   ' + mode[i] + ' ' + str(freq[i]))
	print('Select mode by number (0 for change setting; a for all modes; q for quit)') # prompt for input
	stop = False
	tmp = input('>>> ')
	if tmp.isalpha():
		if tmp == 'q' or tmp == 'Q' :
			stop = True
			qmode = 100000  # number bigger than maxmode
		elif (tmp == 'a') or (tmp == 'A'):
			for i in range(1,imode+1):
				vestaout(i)
			stop = True
			qmode = 100000 # number bigger than maxmode
	else:
	   qmode = int(tmp)
	#qmode = int(input('>>> '))
	if stop: # if q, exit
		exit()
	if qmode == 0: # if 0, change arrow style
		while True:
			print('Select number') # prompt for input
			print('1: change arrow scale? current = ' + str(scale))
			print('2: change arrow radius? current = ' + str(radius))
			print('3: change arrow color? current = ' + arrowRGB)
			i = int(input('>>> '))
			if i > 0 and i< 4:
				if i == 1:
					print('Input new scale')
					scale = float(input('>>> '))
					break
				elif i == 2:
					print('Input new radius')
					radius = float(input('>>> '))
					break
				elif i ==3:
					print('Input new RGB (255 0 0)')
					arrowRGB = input('>>> ')
					break
	if qmode > 0 and qmode < imode+1: # if given mode no. is OK, then proceed
		vestaout(qmode) # output vesta file for a specified mode
		# comment out below and uncomment lines for Windows or Linux
		cmd = "open -a VESTA" + " " + f3 # for Mac
#		cmd = r"C:\Program Files\VESTA-win64\VESTA.exe" + f3 # for Windows
#		cmd = "/usr/bin/VESTA " + f3  + " &" # for Linux, change directory if necessary
		subprocess.call(cmd,shell=True)
